from .grammar import grammar
from .visitor import parse
from .tree import Type
from .unspace import UnspaceVisitor
from typing import Any


__version__ = "0.7"


def Schema(source: str, verbose=False) -> tree.Schema:
    return parse('schema', source)


def Definitions(source: str, verbose=False) -> tree.Definitions:
    return parse('definitions', source)



from abc import ABC, abstractmethod
from typing import NamedTuple, Optional, Set, Tuple
import json
import jsonschema


class Type(ABC):

    CONSTRUCTOR_KWARGS: Tuple[str, ...] = ()

    def __init__(self, **kwargs):
        self._jsonschema = None  # Will be filled as a cache on demand
        for name in self.CONSTRUCTOR_KWARGS:
            setattr(self, name, kwargs[name])

    @abstractmethod
    def to_jsonschema(self):
        pass

    @property
    def jsonschema(self):
        """Cached compilation of the corresponding jsonschema."""
        if self._jsonschema is None:
            self._jsonschema = self.to_jsonschema()
        return self._jsonschema

    def __str__(self):
        a = [k + "=" + str(getattr(self, k)) for k in self.CONSTRUCTOR_KWARGS]
        return f"{self.__class__.__name__}({', '.join(a)})"

    def prettily(self):
        """Return a unicode, pretty-printed representation of me."""
        def indent(text):
            return '\n'.join(('    ' + line) for line in text.splitlines())
        if not self.CONSTRUCTOR_KWARGS:
            return self.__class__.__name__
        else:
            acc = []
            for name in self.CONSTRUCTOR_KWARGS:
                value = getattr(self, name)
                # TODO go through lists and tuples
                r = value.prettily() if isinstance(value, Type) else repr(value)
                acc.append(f"{name} = {r}")
            return self.__class__.__name__ + "(\n" + indent("\n".join(acc)) + "\n)"

    __repr__ = __str__


class Schema(Type):

    CONSTRUCTOR_KWARGS = ("value", "definitions")

    def to_jsonschema(self, check_definitions=True, prune_definitions=True):
        r = self.value.to_jsonschema()
        sdv: Dict[str, Type] = self.definitions.values
        if self.definitions.values:
            r['definitions'] = self.definitions.to_jsonschema()
            if prune_definitions:
                self._prune_definitions(r)
        if check_definitions:
            self._check_definitions(r)
        r["$schema"] = "http://json-schema.org/draft-07/schema#"
        return r

    def _check_definitions(self, schema):
        """Verify that all references have their definition."""
        occurring_references = self._get_references(schema)
        for k in occurring_references:
            if k not in self.definitions.values.keys():
                raise ValueError(f"Missing definition for {k}")

    def _prune_definitions(self, schema):
        """Remove unused definitions. May iterate more than once, because
        some references may be used in other definitions."""
        occurring_references = self._get_references(schema)
        eliminated_references = False
        while True:
            for k in list(schema['definitions'].keys()):
                # Freeze the keys into a list, so that `del` won't interfere
                if k not in occurring_references:
                    del schema['definitions'][k]
                    eliminated_references = True
            if eliminated_references:
                # Some unused references have been removed.
                # By removing their definitions, maybe some other
                # references became unused => try to prune again.
                occurring_references = self._get_references(schema)
                eliminated_references = False
            else:
                # Reached a fix-point, nothing left to eliminate
                return

    def _get_references(self, x) -> Set[str]:
        """Extract every definition usage from a compiled jsonschema,
        so that it can be checked that they are all defined."""
        if isinstance(x, dict):
            if "$ref" in x:
                return {x["$ref"].rsplit("/", 1)[-1]}
            else:
                r = set()
                for y in x.values():
                    r |= self._get_references(y)
                return r
        elif isinstance(x, list):
            return set.union(*(self._get_references(y) for y in x))
        else:
            return set()

    def _combine(self, other, op):
        args = []
        if isinstance(other, Schema):
            # Two schemas: combine main entries and merge defition dicts
            for schema in (self, other):
                content = schema.value
                if isinstance(content, Operator) and content.operator == op:
                    args.extend(content.values)  # Flatten associative operations
                else:
                    args.append(content)
            combined_content = Operator(operator=op, values=args)
            combined_defs = self.definitions | other.definitions
            return Schema(value=combined_content, definitions=combined_defs)
        elif op == 'anyOf' and isinstance(other, Definitions):
            # Schema + additional definitions
            # TODO Maybe '+' is a more appropriate operator than '|'?
            combined_defs = self.definitions | other
            return Schema(value=self.value, definitions=combined_defs)
        else:
            raise ValueError("Schemas can only be combined with each other or with definitions")


    def __and__(self, other):
        return self._combine(other, 'allOf')

    def __or__(self, other):
        return self._combine(other, 'anyOf')

    def validate(self, data=None):
        if data is None:  # Validate the schema itself
            jsonschema.Draft7Validator.check_schema(self.jsonschema)
        else:  # Validate a piece of data against the schema
            jsonschema.validate(
                data,
                self.jsonschema,
                format_checker=jsonschema.draft7_format_checker,
            )


class Definitions(Type):

    CONSTRUCTOR_KWARGS = ("values",)

    def to_jsonschema(self):
        return {k: v.to_jsonschema() for k, v in self.values.items()}

    def __or__(self, other):
        if isinstance(other, Definitions):
            overlap = set(self.values.keys()) & set(other.values.keys())
            if overlap:
                conflicts = ", ".join(sorted(overlap))
                raise ValueError(f"Cannot merge definitions, conflict over {conflicts}")
            defs = dict(self.values)
            defs.update(other.values)
            return Definitions(values=defs)
        elif isinstance(other, Schema):
            return other | self
        else:
            raise ValueError("Cannot perform 'or' on Definitions and that")


class Integer(Type):

    CONSTRUCTOR_KWARGS = ("cardinal", "multiple")

    def to_jsonschema(self):
        (card_min, card_max) = self.cardinal
        r = {"type": "integer"}
        if card_min is not None:
            r["minimum"] = card_min
        if card_max is not None:
            r["maximum"] = card_max
        if self.multiple is not None:
            r["multipleOf"] = self.multiple
        return r


class String(Type):

    CONSTRUCTOR_KWARGS = ("cardinal", "format", "regex")

    def to_jsonschema(self):
        r = {"type": "string"}
        card_min, card_max = self.cardinal
        if card_min is not None:
            r["minLength"] = card_min
        if card_max is not None:
            r["maxLength"] = card_max
        if self.format is not None:
            r["format"] = self.format
        if self.regex is not None:
            r["pattern"] = self.regex
        return r


class Litteral(Type):
    CONSTRUCTOR_KWARGS = ("value",)
    def to_jsonschema(self):
        return {"type": self.value}


class Constant(Type):
    CONSTRUCTOR_KWARGS = ("value",)
    def to_jsonschema(self):
        return {"const": self.value}


class Operator(Type):
    CONSTRUCTOR_KWARGS = ("operator", "values")
    def to_jsonschema(self):
        return {self.operator: [v.to_jsonschema() for v in self.values]}


class Not(Type):
    CONSTRUCTOR_KWARGS = ("value",)
    def to_jsonschema(self):
        return {"not": self.value.to_jsonschema()}


class Enum(Type):
    CONSTRUCTOR_KWARGS = ("values",)
    def to_jsonschema(self):
        return {"enum": list(self.values)}


class Reference(Type):
    CONSTRUCTOR_KWARGS = ("value",)
    def to_jsonschema(self):
        return {"$ref": "#/definitions/" + self.value}


class ObjectProperty(NamedTuple):
    name: Optional[str]
    optional: bool
    type: Type


class Object(Type):

    CONSTRUCTOR_KWARGS = ("properties", "cardinal",
                          "additional_property_types",
                          "additional_property_names")

    def to_jsonschema(self):
        card_min, card_max = self.cardinal
        r = {"type": "object"}
        properties = {}
        required = []
        for (k, opt, v) in self.properties:
            if not opt:
                required.append(k)
            if v is not None:
                properties[k] = v.to_jsonschema()
        if required:
            r["required"] = required
        if properties:
            r["properties"] = properties

        if self.additional_property_types is False:
            r["additionalProperties"] = False
        elif self.additional_property_types is not None:
            r["additionalProperties"] = self.additional_property_types.to_jsonschema()
        if self.additional_property_names is not None:
            r['propertyNames'] = self.additional_property_names.to_jsonschema()

        implicit_card_min = len(required)
        implicit_card_max = (
            len(properties) if r.get("additionalProperties") is False else None
        )
        if card_min is not None:
            if card_min > implicit_card_min:
                r["minProperties"] = card_min
            if implicit_card_max is not None and card_min > implicit_card_max:
                raise ValueError(
                    f"Can only have up to {implicit_card_max} properties, not {card_min}"
                )
        if card_max is not None:
            if implicit_card_max is None or card_max < implicit_card_max:
                r["maxProperties"] = card_max
            if implicit_card_min > card_max:
                raise ValueError(
                    f"Must have at least {implicit_card_min} properties, which is more than {card_max}"
                )
        return r


class Array(Type):

    CONSTRUCTOR_KWARGS = ("items", "additional_items", "cardinal", "unique")

    def to_jsonschema(self):
        #types = self.kwargs["types"]
        #extra_type = self.kwargs["additional_types"]
        #card_min, card_max = self.kwargs["cardinal"]
        r = {"type": "array"}

        if self.items:  # Tuple array
            r["items"] = [item.to_jsonschema() for item in self.items]
            if self.additional_items is False:  # No extra items allowed
                r["additionalItems"] = False
            elif self.additional_items is True:  # Extra items with any type
                pass
            else:  # extra items allowed, but wiht a constrained type
                r["additionalItems"] = self.additional_items.to_jsonschema()
        elif isinstance(self.additional_items, Type):  # List array with homogeneous type
            r["items"] = self.additional_items.to_jsonschema()

        card_min, card_max = self.cardinal
        implicit_card_min = len(self.items)
        implicit_card_max = len(self.items) if self.additional_items is False else None
        if card_min is not None and card_min > implicit_card_min:
            r["minItems"] = card_min
            if implicit_card_max is not None and card_min > implicit_card_max:
                raise ValueError(
                    f"Can only have up to {implicit_card_max} items, not {card_min}"
                )
        if card_max is not None:
            if implicit_card_min > card_max:
                raise ValueError(
                    f"Must have at least {implicit_card_min} items, which is more than {card_max}"
                )
            if implicit_card_max is None or card_max < implicit_card_max:
                r["maxItems"] = card_max

        if self.unique:
            r['uniqueItems'] = True

        return r



import sys
import json
from argparse import ArgumentParser

from .visitor import parse


def main():
    parser = ArgumentParser(description="Convert from a compact DSL into full JSON schema.")
    parser.add_argument('-o', '--output', default="-",
                        help="Output file; defaults to stdout")
    parser.add_argument('-v', '--verbose', action='store_const', const=True, default=False,
                        help="Verbose output")
    parser.add_argument('filename', nargs="?", default="-",
                        help="Input file; use '-' to read from stdin.")
    parser.add_argument('--version', action='store_const', const=True, default=False,
                        help="Display version and exit")
    args = parser.parse_args()

    if args.version:
        from . import __version__
        print(f"JSON Schema compact notation v{__version__}.")
        exit(0)

    input = open(args.filename, 'r') if args.filename != "-" else sys.stdin
    output = open(args.output, 'w') if args.output != "-" else sys.stdout

    source = input.read()
    schema = parse('schema', source, verbose=args.verbose)
    try:
        # TODO Try and guess TTY width
        import jsview
        result = jsview.dumps(schema.jsonschema)
    except ModuleNotFoundError:
        # Raw printing if jsview isn't installed
        result = json.dumps(schema.jsonschema)

    output.write(result+"\n")
    output.flush()



"""PEG Grammar for JSON-Schema compact notation."""
from parsimonious import Grammar

grammar = Grammar(r"""
schema = _ type _ opt_definitions _
type = sequence_and (_ or _ sequence_and)*
sequence_and = simple_type (_ and _ simple_type)*
simple_type = litteral / string / object / integer / array /
       lit_regex / lit_format / constant / parens / not_type / def_reference
parens = lparen _ type _ rparen

litteral = "boolean" / "null" / "number"

lit_integer =  ~"0x[0-9a-fA-F]+" / ~"[0-9]+"
lit_string = ~"\"([^\"\\\\]|\\\\.)*\""
lit_regex = regex_prefix lit_string
lit_format = format_prefix lit_string

constant = ~"`[^`]+`"

# Clearer definition of quoted strings
# lit_string = quote non_quote* quote
# quote = "\""
# non_quote = escaped_character / r"[^\"]"
# escaped_character = r"\\\\."

# Allow quoted strings in JSON constants:
# # 1. the barbaric way
# constant = ~"[^\"]*(\"([^\"\\\\]|\\\\.)*\"[^\"]*)*"
# # 2. the PEGgy way
# constant = backquote (neither_quote_nor_backquote / lit_string)* backquote
# backquote = "`"
# neither_quote_nor_backquote = escaped_character / r"[^\"`]"

string = "string" _ opt_cardinal
integer = "integer" _ opt_cardinal _ opt_multiple
opt_multiple = ("/" _ lit_integer)?

not_type = not _ simple_type

regex_prefix = "r"
format_prefix = "f"
wildcard = "_"
or = "|"
and = "&"
not = "not"
lparen = "("
rparen = ")"
lbrace = "{"
rbrace = "}"
lbracket = "["
rbracket = "]"
comma = ","
colon = ":"
question = "?"
star = "*"
plus = "+"
kw_array = "array"
kw_object = "object"
only = "only"
unique = "unique"

opt_cardinal = (lbrace _ card_content _ rbrace)?
card_content = card_2 / card_min / card_max / card_1
card_2 = lit_integer _ comma _ lit_integer
card_1 = lit_integer
card_min = lit_integer _ comma? _ wildcard
card_max = wildcard _ comma? _ lit_integer

object = object_empty / object_non_empty / object_keyword
object_empty = lbrace _ object_only _ rbrace opt_cardinal
object_keyword = kw_object opt_cardinal
object_non_empty = lbrace _
                   object_only _
                   object_pair (_ comma _ object_pair)* _
                   rbrace
                   opt_cardinal
object_only = (only _ ((lit_regex/def_reference/wildcard) _ (colon _ type)?)? comma?)?
object_pair = object_pair_name _ question? _ colon _ object_pair_type
object_pair_name = lit_string / object_pair_unquoted_name
object_pair_unquoted_name = ~"[A-Za-z0-9][-_A-Za-z0-9]*"
object_pair_type = type / wildcard

array = array_empty / array_non_empty
array_empty = ((lbracket _ rbracket) / kw_array) _ opt_cardinal
array_non_empty = lbracket _ array_prefix _
                  type (_ comma _ type)* _
                  array_extra _
                  rbracket _
                  opt_cardinal
array_prefix = ((only / unique) _) *
array_extra = (plus / star)?

opt_definitions = (def_where _ definitions)?
definitions = _ definition (_ def_and _ definition)* _
definition = def_identifier _ def_equal _ type
def_where = "where"
def_and = "and"
def_equal = "="
def_reference = "<" def_identifier ">"
def_identifier = ~"[A-Za-z_][-A-Za-z_0-9]*"

_ = meaninglessness*
meaninglessness = ~r"\s+" / comment
comment = ~r"#[^\r\n]*"
"""
)



from parsimonious import NodeVisitor

class UnspaceVisitor(NodeVisitor):
    """Remove space tokens from a Parsimonious parse tree."""

    WHITESPACE_TOKEN = object()  # Sentinel value

    def visit__(self, node, children) -> object:
        """Replace whitspaces with an easy-to-filter sentinel value."""
        return self.WHITESPACE_TOKEN

    def generic_visit(self, node, children) -> tuple:
        """ The generic visit method. """
        node.children = [x for x in children if x is not self.WHITESPACE_TOKEN]
        return node



import unittest
from . import Schema, Definitions
import json
import jsonschema


class TestJSCN(unittest.TestCase):

    def cmp(self, src: str,sch: dict) -> None:
        sch2 = Schema(src).to_jsonschema()
        del sch2['$schema']
        self.assertDictEqual(sch2, sch)

    def test_simple(self):
        self.cmp("boolean", {"type": "boolean"})
        self.cmp("string", {"type": "string"})
        self.cmp("integer", {"type": "integer"})
        self.cmp("number", {"type": "number"})
        self.cmp("null", {"type": "null"})

    def test_lit_string(self):
        self.cmp('r"foo"', {"type": "string", "pattern": "foo"})
        self.cmp('f"foo"', {"type": "string", "format": "foo"})
        self.cmp('r"foo\\\"bar"', {"type": "string", "pattern": r'foo\"bar'})
        self.cmp('r"foo\\""', {"type": "string", "pattern": r'foo\"'})

    def test_integer(self):
        self.cmp('integer{_,1}', {"type": "integer", "maximum": 1})
        self.cmp('integer{1,_}', {"type": "integer", "minimum": 1})
        self.cmp('integer{0,10}', {"type": "integer", "minimum": 0, "maximum": 10})
        self.cmp('integer/5', {"type": "integer", "multipleOf": 5})

    def test_string(self):
        self.cmp('string{_,1}', {"type": "string", "maxLength": 1})
        self.cmp('string{1,_}', {"type": "string", "minLength": 1})
        self.cmp('string{0,10}', {"type": "string", "minLength": 0, "maxLength": 10})

    def test_constant(self):
        self.cmp("`123`", {"const": 123})
        self.cmp('`"123"`', {"const": "123"})
        self.cmp('`{"a": 1}`', {"const": {"a": 1}})

    def test_enum(self):
        self.cmp("`1`|`2`|`3`", {"enum": [1, 2, 3]})

    def test_object_empty(self):
        obj = {"type": "object"}
        self.cmp("{}", obj)
        self.cmp("object", obj)

    def test_object_simple(self):
        self.cmp("{only foo: integer}",
                 {"type": "object",
                  "required": ["foo"],
                  "properties":
                  {"foo": {"type": "integer"}},
                  "additionalProperties": False})
        self.cmp("{foo: integer}",
                 {"type": "object",
                  "required": ["foo"],
                  "properties":
                  {"foo": {"type": "integer"}}})
        self.cmp("{foo?: integer}",
                 {"type": "object",
                  "properties":
                  {"foo": {"type": "integer"}}})
        self.cmp('{"foo"?: integer}',
                 {"type": "object",
                  "properties":
                  {"foo": {"type": "integer"}}})
        self.cmp('{_: integer}',
                 {"type": "object",
                  "additionalProperties": {"type": "integer"}})

    def test_object_card(self):
        pass  # TODO

    def test_array_empty(self):
        array = {"type": "array"}
        self.cmp("[]", array)
        self.cmp("array", array)

    def test_object_simple(self):
        integer = {"type": "integer"}
        self.cmp("[only integer]",
                 {"type": "array",
                  "items": [integer],
                  "additionalItems": False})
        self.cmp("[integer]",
                 {"type": "array",
                  "items": [integer]})
        self.cmp("[integer*]",
                 {"type": "array",
                  "items": integer}) # List-notation rather than tuple notation
        self.cmp("[integer+]",  # Maybe encoded either that way or with a cardinal
                 {"type": "array",
                  "items": [integer],
                  "additionalItems": integer})

    def test_array_card(self):
        pass  # TODO

    def test_where(self):
        s = '{only <id>: <byte>} where id = r"[a-z]+" and byte = integer{0,0xFF}'
        self.cmp(s, {
            'additionalProperties': {'$ref': '#/definitions/byte'},
            'definitions': {
                'byte': {'maximum': 255, 'minimum': 0, 'type': 'integer'},
                'id': {'pattern': '[a-z]+', 'type': 'string'}},
            'propertyNames': {'$ref': '#/definitions/id'},
            'type': 'object'
        })

    def test_xxx(self):
        Schema(r"""{
            kind: `"aircraft"`,
            mission: string
        } | {
            kind: `"mission"`,
            name: string,
            fleet: {only <id>: string}
        } where id = r"[a-z]+" """).to_jsonschema()

    def test_yyy(self):
        Schema("""
        { instance: <plid>,
          ground: <plid>,
          mission: <plid>,
          fleet: { only <plid>: <aircraft> }
        }
        where plid = r"[A-Z0-9]{4}"
        and aircraft = {
          color?: string,
          status?: string  # TODO bug in JSCN `"online"` | `"offline"`
        }
        """).to_jsonschema()

    def test_prune(self):
        s = Schema("""
        {prop: <used_1>}
        where used_1 = [<used_2>+]
        and unused_1 = [<unused_2>+]
        and used_2 = integer
        and unused_2 = string
        """)
        self.assertSetEqual(
            set(s.jsonschema['definitions'].keys()),
            {'used_1', 'used_2'}
        )

    def test_combine_1(self):
        s = Schema("""{prop: <used_1>}""")
        d = Definitions("""
        used_1 = [<used_2>+]
        and unused_1 = [<unused_2>+]
        and used_2 = integer
        and unused_2 = string
        """)
        s |= d
        self.assertSetEqual(
            set(s.jsonschema['definitions'].keys()),
            {'used_1', 'used_2'}
        )

    def test_combine_2(self):
        s = Schema("{prop: <used_1>}") | \
            Definitions("used_1 = [<used_2>+]") | \
            Definitions("unused_1 = [<unused_2>+]") | \
            Definitions("used_2 = integer") | \
            Definitions("unused_2 = string")
        self.assertSetEqual(
            set(s.jsonschema['definitions'].keys()),
            {'used_1', 'used_2'}
        )

    def test_combine_3(self):
        """combine defs together before adding the result to the schema."""
        s = Schema("{prop: <used_1>}") | (
            Definitions("used_1 = [<used_2>+]") |
            Definitions("unused_1 = [<unused_2>+]") |
            Definitions("used_2 = integer") |
            Definitions("unused_2 = string"))
        self.assertSetEqual(
            set(s.jsonschema['definitions'].keys()),
            {'used_1', 'used_2'}
        )

    def test_combine_schemas(self):
        self.cmp('{foo?: integer} & {bar?: string}',
                 {'allOf': [
                     {'properties': {'foo': {'type': 'integer'}},
                      'type': 'object'},
                     {'properties': {'bar': {'type': 'string'}},
                      'type': 'object'}]})


    def test_missing_def(self):
        with self.assertRaisesRegex(ValueError, "Missing definition"):
            Schema('{foo: <bar>}').to_jsonschema()

    def test_validate(self):
        s = Schema("{only x: integer}")
        s.validate()
        s.validate({"x": 1})
        with self.assertRaises(jsonschema.ValidationError):
            s.validate({"x": "1"})
        with self.assertRaises(jsonschema.ValidationError):
            s.validate({"x": 1, "y": 2})


    def test_def_conflict(self):
        with self.assertRaisesRegex(ValueError, "conflict"):
            Definitions("x = integer") | Definitions("x = string")


if __name__ == '__main__':
    unittest.main()
